/*
    seolizer - a web server log analysis program

    Copyright (C) 1997-2001  Bradford L. Barrett (brad@mrunix.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

    This software uses the gd graphics library, which is copyright by
    Quest Protein Database Center, Cold Spring Harbor Labs.  Please
    see the documentation supplied with the library for additional
    information and license terms, or visit www.boutell.com/gd/ for the
    most recent version of the library and supporting documentation.
*/

/*********************************************/
/* STANDARD INCLUDES                         */
/*********************************************/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>                           /* normal stuff             */
#include <ctype.h>
#include <sys/utsname.h>
#include <sys/times.h>

/* ensure getopt */
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

/* ensure sys/types */
#ifndef _SYS_TYPES_H
#include <sys/types.h>
#endif

/* some systems need this */
#ifdef HAVE_MATH_H
#include <math.h>
#endif

/* SunOS 4.x Fix */
#ifndef CLK_TCK
#define CLK_TCK _SC_CLK_TCK
#endif

#include "linklist.h"
#include "hashtab.h"

extern int verbose;
extern int debug_mode;


#define msg_big_one "msgbigone"
/* internal function prototypes */

SNODEPTR new_snode(char *, int count, char*pageurl);                   /* new search string..      */
u_long   hash(char *);                        /* hash function            */

/* local data */

PNODEPTR pg_htab[MAXHASH];                    /* page table      */
SNODEPTR sr_htab[MAXHASH];                    /* search string table      */
SNODEPTR kw_htab[MAXHASH];                    /* search keywords table      */


void del_htabs()
{
   del_slist(sr_htab);
   del_slist(kw_htab);
   del_plist(pg_htab);
}

/*********************************************/
/* NEW_SNODE - Search str node creation      */
/*********************************************/
SNODEPTR new_snode(char *str, int count, char *pageurl)
{
   SNODEPTR newptr;
   char     *sptr;
   int i;


   if (strlen(str) >= MAXSRCHH)
   {
      if (verbose)
      {
         fprintf(stderr,"[new_snode] %s (%d)",msg_big_one,strlen(str));
         if (debug_mode)
            fprintf(stderr,":\n--> %s",str);
         fprintf(stderr,"\n");
      }
      str[MAXSRCHH-1]=0;
   }

   if ( (sptr=malloc(strlen(str)+1))==NULL ) return (SNODEPTR)NULL;
   strcpy(sptr,str);

   if (( newptr = malloc(sizeof(struct snode))) != NULL)
   {
      newptr->string= sptr;
      newptr->count = count;
      for (i=0;i<MAXHASH;i++)
         newptr->pages[i] = NULL;
      if (pageurl != NULL)
           put_pnode(pageurl,count,newptr->pages, NULL);

   }
   else free(sptr);

   return newptr;
}

/*********************************************/
/* NEW_PNODE - Search pageurl node creation      */
/*********************************************/
PNODEPTR new_pnode(char *pageurl, int count, char *str)
{
   PNODEPTR newptr;
   char     *sptr;
   int i;


   if (strlen(pageurl) >= MAXURL)
   {
      if (verbose)
      {
         fprintf(stderr,"[new_pnode] %s (%d)",msg_big_one,strlen(pageurl));
         if (debug_mode)
            fprintf(stderr,":\n--> %s",pageurl);
         fprintf(stderr,"\n");
      }
      pageurl[MAXURL-1]=0;
   }

   if ( (sptr=malloc(strlen(pageurl)+1))==NULL ) return (PNODEPTR)NULL;
   strcpy(sptr,pageurl);

   if (( newptr = malloc(sizeof(struct pnode))) != NULL)
   {
      newptr->pageurl= sptr;
      newptr->count = 1;
      for (i=0;i<MAXHASH;i++)
         newptr->searches[i] = NULL;
      if (str != NULL)
           put_snode(str,count,newptr->searches, NULL);

   }
   else free(sptr);

   return newptr;
}



/*********************************************/
/* PUT_SNODE - insert/update search str node */
/*********************************************/

int put_snode(char *str, u_long count, SNODEPTR *htab, char *pageurl)
{
   SNODEPTR cptr,nptr;

   if (str[0]==0 || str[0]==' ') return 0;     /* skip bad search strs */

   /* check if hashed */
   if ( (cptr = htab[hash(str)]) == NULL)
   {
      /* not hashed */
      if ( (nptr=new_snode(str, count, pageurl)) != NULL)
      {
         nptr->next = NULL;
         htab[hash(str)] = nptr;
      }
   }
   else
   {
      /* hashed */
      while (cptr != NULL)
      {
         if (strcmp(cptr->string,str)==0)
         {
            /* found... bump counter */
            if (pageurl != NULL)
               put_pnode(pageurl,count,cptr->pages, NULL);
            cptr->count+=count;
            return 0;
         }
         cptr = cptr->next;
      }
      /* not found... */
      if ( (nptr = new_snode(str, count, pageurl)) != NULL)
      {
         nptr->next  = htab[hash(str)];
         htab[hash(str)]=nptr;
      }
   }
   return nptr==NULL;
}

/*********************************************/
/* PUT_PNODE - insert/update page node */
/*********************************************/

int put_pnode(char *pageurl, u_long count, PNODEPTR *htab, char *searchstr)
{
   PNODEPTR cptr,nptr;

   if (pageurl[0]==0 || pageurl[0]==' ') return 0;     /* skip bad search strs */

   /* check if hashed */
   if ( (cptr = htab[hash(pageurl)]) == NULL)
   {
      /* not hashed */
      if ( (nptr=new_pnode(pageurl, count, searchstr)) != NULL)
      {
         nptr->count = count;
         nptr->next = NULL;
         htab[hash(pageurl)] = nptr;
      }
   }
   else
   {
      /* hashed */
      while (cptr != NULL)
      {
         if (strcmp(cptr->pageurl,pageurl)==0)
         {
            /* found... bump counter */
            if (searchstr != NULL)
               put_snode(searchstr,count,cptr->searches, NULL);

            cptr->count+=count;
            return 0;
         }
         cptr = cptr->next;
      }
      /* not found... */
      if ( (nptr = new_pnode(pageurl, count, searchstr)) != NULL)
      {
         nptr->count = count;
         nptr->next  = htab[hash(pageurl)];
         htab[hash(pageurl)]=nptr;
      }
   }
   return nptr==NULL;
}



/*********************************************/
/* DEL_SLIST - delete search str hash table  */
/*********************************************/

void	del_slist(SNODEPTR *htab)
{
   /* free memory used by hash table */
   SNODEPTR aptr,temp;
   int i;

   for (i=0;i<MAXHASH;i++)
   {
      if (htab[i] != NULL)
      {
         aptr = htab[i];
         while (aptr != NULL)
         {
            temp = aptr->next;
            free (aptr->string);
            free (aptr);
            aptr = temp;
         }
         htab[i]=NULL;
      }
   }
}

/*********************************************/
/* DEL_PLIST - delete */
/*********************************************/

void	del_plist(PNODEPTR *htab)
{
   /* free memory used by hash table */
   PNODEPTR aptr,temp;
   int i;

   for (i=0;i<MAXHASH;i++)
   {
      if (htab[i] != NULL)
      {
         aptr = htab[i];
         while (aptr != NULL)
         {
            temp = aptr->next;
            free (aptr->pageurl);
            free (aptr);
            aptr = temp;
         }
         htab[i]=NULL;
      }
   }
}


/*********************************************/
/* HASH - return hash value for string       */
/*********************************************/

u_long hash(char *str)
{
   u_long hashval;
   for (hashval = 0; *str != '\0'; str++)
      hashval = *str + 31 * hashval;
   return hashval % MAXHASH;
}

/*********************************************/
/* LOAD_SRCH_ARRAY - load up the sort array  */
/*********************************************/
u_long load_srch_array(SNODEPTR *sr_htab, SNODEPTR *pointer)
{
   SNODEPTR sptr;
   int      i;
   u_long   ctr = 0;

   /* load the array */
   for (i=0;i<MAXHASH;i++)
   {
      sptr=sr_htab[i];
      while (sptr!=NULL)
      {
         if (pointer==NULL) ctr++;       /* fancy way to just count 'em    */
         else *(pointer+ctr++)=sptr;     /* otherwise, really do the load  */
         sptr=sptr->next;
      }
   }
   return ctr;   /* return number loaded */
}

/*********************************************/
/* LOAD_PAGE_ARRAY - load up the sort array  */
/*********************************************/
u_long load_page_array(PNODEPTR *page_htab, PNODEPTR *pointer)
{
   PNODEPTR sptr;
   int      i;
   u_long   ctr = 0;

   /* load the array */
   for (i=0;i<MAXHASH;i++)
   {
      sptr=page_htab[i];
      while (sptr!=NULL)
      {
         if (pointer==NULL) ctr++;       /* fancy way to just count 'em    */
         else *(pointer+ctr++)=sptr;     /* otherwise, really do the load  */
         sptr=sptr->next;
      }
   }
   return ctr;   /* return number loaded */
}

